use std::fmt;
use std::future::Future;
use std::pin::Pin;

use camino::Utf8Path;
use camino::Utf8PathBuf;
use libunftp::auth::AuthenticationError;
use libunftp::auth::Credentials;
use libunftp::auth::UserDetail;
use tokio::fs::File;
use tokio::io::AsyncRead;

use yadms_storage::document::Document;
use yadms_storage::Error as InternalError;
use yadms_storage::tag::Tag;
use yadms_storage::Session as YadmsSession;

use crate::error::Error;

#[derive(Debug)]
pub struct Session {
	inner: YadmsSession
}

impl Session {
	pub fn get_root(&self) -> &Utf8Path {
		self.inner.get_root()
	}

	pub fn canonicalize_path(&self, path: &Utf8Path) -> Result<Utf8PathBuf, Error> {
		Ok(self.inner.canonicalize_path(path)?)
	}

	pub async fn get_tag_by_canonical_path(&self, path: &Utf8Path) -> Result<Tag, Error> {
		Ok(self.inner.get_tag_by_canonical_path(path).await?)
	}

	pub async fn get_document_by_tag_id_and_name(&self, tag_id: u32, name: &Utf8Path) -> Result<Document, Error> {
		Ok(self.inner.get_document_by_tag_id_and_name(tag_id, name).await?)
	}

	pub async fn get_document_by_path(&self, path: &Utf8Path) -> Result<Document, Error> {
		Ok(self.inner.get_document_by_path(path).await?)
	}

	pub async fn stat(&self, path: &Utf8Path) -> Result<yadms_storage::PathType, Error> {
		Ok(self.inner.stat(path).await?)
	}

	pub async fn get_tags(&self, parent_id: Option<u32>, include_descendents: bool) -> Result<Vec<Tag>, Error> {
		Ok(self.inner.get_tags(parent_id, include_descendents).await?)
	}

	pub async fn get_documents_by_tag_id(&self, tag_id: u32) -> Result<Vec<Document>, Error> {
		Ok(self.inner.get_documents_by_tag_id(tag_id).await?)
	}

	pub async fn open_document(&self, id: u32) -> Result<(File, Utf8PathBuf), Error> {
		Ok(self.inner.open_document(id).await?)
	}

	pub async fn create_document(&self, path: Utf8PathBuf, data: impl AsyncRead + Unpin, allow_overwrite: bool) -> Result<Document, Error> {
		Ok(self.inner.create_document(path, data, allow_overwrite).await?)
	}

	pub async fn delete_document(&self, id: u32) -> Result<(), Error> {
		Ok(self.inner.delete_document(id).await?)
	}

	pub async fn untag_document(&self, document_id: u32, tag_id: u32) -> Result<Document, Error> {
		Ok(self.inner.untag_document(document_id, tag_id).await?)
	}

	pub async fn mkdir(&self, path: &Utf8Path) -> Result<(), Error> {
		Ok(self.inner.mkdir(path).await?)
	}

	pub async fn rmdir(&self, path: &Utf8Path) -> Result<(), Error> {
		Ok(self.inner.rmdir(path).await?)
	}

	pub async fn create_tag(&self, parent_id: Option<u32>, name: &Utf8Path) -> Result<Tag, Error> {
		Ok(self.inner.create_tag(parent_id, name.as_str()).await?)
	}

	pub async fn delete_tag(&self, id: u32) -> Result<(), Error> {
		Ok(self.inner.delete_tag(id).await?)
	}

	pub async fn move_dir(&self, source: &Utf8Path, destination: &Utf8Path) -> Result<(), Error> {
		Ok(self.inner.move_dir(source, destination).await?)
	}

	pub async fn move_document(&self, document_id: u32, new_path: &Utf8Path) -> Result<Document, Error> {
		Ok(self.inner.move_document(document_id, new_path).await?)
	}
}

impl From<YadmsSession> for Session {
	fn from(inner: YadmsSession) -> Self {
		Self{inner}
	}
}

impl fmt::Display for Session {
	fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
		self.inner.fmt(f)
	}
}

impl UserDetail for Session {
	fn account_enabled(&self) -> bool {
		true
	}
}

#[derive(Debug)]
pub struct Authenticator {
	store: yadms_storage::DocumentStore
}

impl Authenticator {
	pub(crate) async fn new() -> std::result::Result<Self, Box<dyn std::error::Error>> {
		Ok(Self{
			store: yadms_storage::DocumentStore::new().await?
		})
	}

	async fn _authenticate(&self, username: &str, creds: &Credentials) -> core::result::Result<Session, AuthenticationError> {
		let password = creds.password.as_ref().ok_or_else(|| AuthenticationError::new("Password required"))?;
		match self.store.authenticate("yadms-ftp-server", username, &password).await {
			Ok(v) => Ok(v.into()),
			Err(e) => match e {
				InternalError::LoginFailed => Err(AuthenticationError::BadPassword),
				e => Err(AuthenticationError::new(e.to_string()))
			}
		}
	}
}

impl libunftp::auth::Authenticator<Session> for Authenticator {
	fn authenticate<'this, 'un, 'creds, 'async_trait>(&'this self, username: &'un str, creds: &'creds Credentials) -> Pin<Box<dyn Future<Output = core::result::Result<Session, AuthenticationError>> + Send + 'async_trait>>
	where
		'this: 'async_trait,
		'un: 'async_trait,
		'creds: 'async_trait,
		Self: 'async_trait
	{
		Box::pin(self._authenticate(username, creds))
	}

	fn name(&self) -> &str {
		"YADMS"
	}
}

