use std::fmt;

use libunftp::storage::Error as FtpError;
use libunftp::storage::ErrorKind;

use yadms_storage::Error as InternalError;

#[derive(Debug)]
#[repr(transparent)]
pub struct Error(InternalError);

impl fmt::Display for Error {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
		self.0.fmt(f)
	}
}

impl Error {
	pub fn inner(&self) -> &InternalError {
		&self.0
	}
}

impl From<InternalError> for Error {
	fn from(inner: InternalError) -> Self {
		Self(inner)
	}
}

impl From<Error> for FtpError {
	fn from(this: Error) -> Self {
		// Some of what _seem like_ they should be "bad request" type errors
		//    are actually internal; because the client is presented with a
		//    filesystem, these error cases exist in the boundary interface
		//    between YADMS core and the FTP interface.  This means that - to
		//    a client - they're internal.  For example,
		//    `InternalError::DocumentIdNotFound` represents a logic error in
		//    this crate, because it somehow successfully looked up a document
		//    by path, but then tried and failed to retrieve it by ID.
		let kind = match this.0 {
			InternalError::ParseInt(_) => ErrorKind::LocalError,
			InternalError::LoginFailed => ErrorKind::PermissionDenied,
			InternalError::SessionNotFound => ErrorKind::LocalError,
			InternalError::PathNotFound(_) => ErrorKind::PermanentFileNotAvailable,
			InternalError::PathIsDirectory(_) => ErrorKind::PermissionDenied,
			InternalError::PathIsFile(_) => ErrorKind::PermissionDenied,
			InternalError::AlreadyExists(_) => ErrorKind::PermanentFileNotAvailable,
			InternalError::TagContinuity => ErrorKind::LocalError,
			InternalError::DocumentIdNotFound(_) => ErrorKind::LocalError,
			InternalError::DocumentNameNotFound(_) => ErrorKind::PermanentFileNotAvailable,
			InternalError::TagIdNotFound(_) => ErrorKind::LocalError,
			InternalError::Database(_) => ErrorKind::LocalError,
			InternalError::IO(_) => ErrorKind::LocalError,
			InternalError::WrongHashSize => ErrorKind::LocalError,
			InternalError::Rcrypt(_) => ErrorKind::LocalError,
			InternalError::InvalidPath => ErrorKind::FileNameNotAllowedError
		};
		FtpError::new(kind, this.0.user_facing_message())
	}
}

