extern crate libunftp;
extern crate tokio;

extern crate yadms_storage;

use std::convert::TryFrom;
use std::convert::TryInto;
use std::fmt::Debug;
use std::path::Path;
use std::path::PathBuf;
use std::time::SystemTime;

use camino::Utf8Path;
use camino::Utf8PathBuf;
use libunftp::storage::Error;
use libunftp::storage::ErrorKind;
use libunftp::storage::Fileinfo;
use libunftp::storage::Result;
use libunftp::storage::StorageBackend;
use tokio::io::AsyncRead;
use tokio::io::AsyncSeekExt;

use yadms_storage::PathType;

use crate::auth::Session;

pub(crate) mod error;
pub mod metadata;
pub(crate) mod path;

#[derive(Debug)]
pub(crate) struct Backend;

// Similar to libunftp implementation:
//    https://docs.rs/libunftp/0.10.0/src/libunftp/storage/filesystem.rs.html#41
fn full_path(path: &Path) -> Result<Utf8PathBuf> /* {{{ */ {
	// This is super janky but Brother's ADS-2700W is ignoring me when I tell
	//    it to just put stuff in /documents; it's trying to CWD ~
	let mut path = Utf8PathBuf::try_from(path.to_path_buf())
		.map_err(|e| Error::new(ErrorKind::FileNameNotAllowedError, e))?
		.into_string();
	path = path.replace("~", "/");
	path = path.replace("//", "/");
	let path = match path.starts_with('/') {
		true => Utf8PathBuf::from(path),
		false => Utf8PathBuf::from("/").join(path)
	};
	Ok(yadms_storage::canonicalize_path(&path).ok_or_else(|| Error::from(ErrorKind::FileNameNotAllowedError))?)
} // }}}

fn require_absolute_document_path(path: &Path) -> Result<Utf8PathBuf> /* {{{ */ {
	let mut path = full_path(path)?;
	path = match path.strip_prefix("/documents") {
		Ok(v) => v.to_path_buf(),
		Err(_) => return Err(Error::from(ErrorKind::FileNameNotAllowedError))
	};
	Ok(Utf8PathBuf::from("/").join(path))
} // }}}

impl Backend {
	pub(crate) async fn new() -> std::result::Result<Self, Box<dyn std::error::Error>> /* {{{ */ {
		Ok(Self)
	} // }}}

	async fn parse_path<P: AsRef<Path>>(&self, user: &Session, path: P) -> Result<path::Path> /* {{{ */ {
		let absolute_path = full_path(path.as_ref())?;
		if(absolute_path == "/") {
			return Ok(path::Path::new(absolute_path.clone(), "/..".into(), path::PathType::Root));
		} else if let Ok(path) = absolute_path.strip_prefix("/tags") {
			if(path == "") {
				return Ok(path::Path::new(absolute_path.clone(), "/".into(), path::PathType::Tag(None)));
			}
			let path = Utf8PathBuf::from("/").join(path);
			match user.get_tag_by_canonical_path(&path).await {
				Ok(tag) => return Ok(path::Path::new(absolute_path.clone(), path.to_path_buf(), path::PathType::Tag(Some(tag)))),
				Err(e) if e.inner().is_not_found() => (),
				Err(_) => return Err(Error::from(ErrorKind::LocalError))
			};
			let tag_path = absolute_path.parent().expect("tag parent() fail");
			match user.get_tag_by_canonical_path(tag_path).await {
				Ok(tag) => {
					let filename = absolute_path.file_name().expect("tag file_name() fail");
					match user.get_document_by_tag_id_and_name(tag.get_id(), filename.as_ref()).await {
						Ok(doc) => return Ok(path::Path::new(absolute_path.clone(), path.to_path_buf(), path::PathType::Document(Some(doc)))),
						Err(e) if e.inner().is_not_found() => return Ok(path::Path::new(absolute_path.clone(), path.to_path_buf(), path::PathType::Tag(None))),
						Err(_) => return Err(Error::from(ErrorKind::LocalError))
					};
				},
				Err(e) if e.inner().is_not_found() => return Ok(path::Path::new(absolute_path.clone(), path.to_path_buf(), path::PathType::Tag(None))),
				Err(_) => return Err(Error::from(ErrorKind::LocalError))
			};
		} else if let Ok(path) = absolute_path.strip_prefix("/documents") {
			let path = Utf8PathBuf::from("/").join(path);
			match user.get_document_by_path(&path).await {
				Ok(doc) => return Ok(path::Path::new(absolute_path.clone(), path.to_path_buf(), path::PathType::Document(Some(doc)))),
				Err(e) if e.inner().is_not_found() => (),
				Err(_) => return Err(Error::from(ErrorKind::LocalError))
			};
			let is_dir = match user.stat(&path).await {
				Ok(v) => v == PathType::Directory,
				Err(_) => return Err(Error::from(ErrorKind::LocalError))
			};
			if(is_dir) {
				return Ok(path::Path::new(absolute_path.clone(), path.to_path_buf(), path::PathType::Directory));
			}
			return Ok(path::Path::new(absolute_path.clone(), path.to_path_buf(), path::PathType::Document(None)));
		}
		return Err(Error::from(ErrorKind::PermanentFileNotAvailable));
	} // }}}
}

#[async_trait]
impl StorageBackend<Session> for Backend {
	type Metadata = metadata::Metadata;

	async fn metadata<P: AsRef<Path> + Send>(&self, user: &Session, path: P) -> Result<Self::Metadata> /* {{{ */ {
		let path = self.parse_path(user, path).await?;
		let full_path = user.canonicalize_path(&path.path)?;
		Ok(tokio::fs::symlink_metadata(full_path).await?.into())
	} // }}}

	async fn list<P>(&self, user: &Session, path: P) -> Result<Vec<Fileinfo<PathBuf, Self::Metadata>>> /* {{{ */
	where
		P: AsRef<Path> + Send,
		<Self as StorageBackend<Session>>::Metadata: libunftp::storage::Metadata
	{
		let path = self.parse_path(user, path).await?;
		match path.contents {
			path::PathType::Root => Ok(vec![
				Fileinfo{
					path: "documents".into(),
					metadata: metadata::Metadata::new(0, true, false, false, SystemTime::now(), 1000, 1000)
				},
				Fileinfo{
					path: "tags".into(),
					metadata: metadata::Metadata::new(0, true, false, false, SystemTime::now(), 1000, 1000)
				}
			]),
			path::PathType::Document(Some(_)) => Ok(vec![Fileinfo{
				path: path.path.into(),
				metadata: tokio::fs::symlink_metadata(user.canonicalize_path(&path.internal_path)?).await?.into()
			}]),
			path::PathType::Document(None) => Err(Error::from(ErrorKind::PermanentFileNotAvailable)),
			path::PathType::Directory => {
				let full_path = user.canonicalize_path(&path.internal_path)?;
				let root_path = user.get_root();
				let mut dir = match tokio::fs::read_dir(full_path).await {
					Ok(dir) => dir,
					Err(e) => return Err(Error::from(match e.kind() {
						tokio::io::ErrorKind::NotFound => ErrorKind::PermanentFileNotAvailable,
						tokio::io::ErrorKind::PermissionDenied => ErrorKind::PermissionDenied,
						_ => ErrorKind::LocalError
					}))
				};
				let mut entries = Vec::new();
				while let Ok(Some(entry)) = dir.next_entry().await {
					let path = entry.path();
					let relative_path = path.strip_prefix(root_path).expect("list strip_prefix() fail");
					entries.push(Fileinfo{
						path: relative_path.into(),
						metadata: tokio::fs::symlink_metadata(path).await?.into()
					});
				}
				Ok(entries)
			},
			path::PathType::Tag(tag) => {
				let tags = user.get_tags(tag.as_ref().map(|tag| tag.get_id()), false).await?;
				let documents = match tag {
					Some(tag) => user.get_documents_by_tag_id(tag.get_id()).await?,
					None => Vec::new()
				};
				let mut entries = Vec::with_capacity(tags.len() + documents.len());
				for tag in tags.iter() {
					entries.push(Fileinfo{
						path: tag.get_name().into(),
						metadata: metadata::Metadata::new(0, true, false, false, SystemTime::now(), 1000, 1000)
					});
				}
				for doc in documents.iter() {
					let janky_path = Utf8PathBuf::from(format!(
						"{} -> {}",
						doc.get_path().file_name().ok_or_else(|| Error::from(ErrorKind::FileNameNotAllowedError))?,
						doc.get_path()
					));
					println!("janky_path: {}", janky_path);
					entries.push(Fileinfo{
						path: janky_path.into(),
						metadata: metadata::Metadata::new(0, false, true, true, SystemTime::now(), 1000, 1000)
					});
				}
				Ok(entries)
			}
		}
	} // }}}

	async fn get<P: AsRef<Path> + Send>(&self, user: &Session, path: P, start_pos: u64) -> Result<Box<dyn tokio::io::AsyncRead + Send + Sync + Unpin>> /* {{{ */ {
		let path = self.parse_path(user, path).await?;
		let mut fd = match path.contents {
			path::PathType::Root => return Err(Error::from(ErrorKind::PermissionDenied)),
			path::PathType::Directory => return Err(Error::from(ErrorKind::PermissionDenied)),
			path::PathType::Document(Some(doc)) => user.open_document(doc.get_id()).await?.0,
			path::PathType::Document(None) => return Err(Error::from(ErrorKind::PermanentFileNotAvailable)),
			path::PathType::Tag(_) => return Err(Error::from(ErrorKind::PermissionDenied))
		};
		if(start_pos > 0) {
			fd.seek(std::io::SeekFrom::Start(start_pos)).await?;
		}
		Ok(Box::new(fd))
	} // }}}

	async fn put<P: AsRef<Path> + Send, R: AsyncRead + Send + Sync + 'static + Unpin>(&self, user: &Session, bytes: R, path: P, start_pos: u64) -> Result<u64> /* {{{ */ {
		if(start_pos != 0) {
			eprintln!("put():  Not able to seek files at this time");
			return Err(Error::from(ErrorKind::LocalError));
		}
		let path = require_absolute_document_path(path.as_ref())?;
		match user.create_document(path, bytes, true).await {
			Ok(doc) => Ok(doc.get_size() as u64),
			Err(e) => {
				println!("Error from create_document: {:?}", e);
				Err(e.into())
			}
		}
	} // }}}

	async fn del<P: AsRef<Path> + Send>(&self, user: &Session, path: P) -> Result<()> /* {{{ */ {
		let path = self.parse_path(user, path).await?;
		let doc = match &path.contents {
			path::PathType::Root => return Err(Error::from(ErrorKind::PermissionDenied)),
			path::PathType::Directory => return Err(Error::from(ErrorKind::PermissionDenied)),
			path::PathType::Document(Some(doc)) => doc,
			path::PathType::Document(None) => return Err(Error::from(ErrorKind::PermanentFileNotAvailable)),
			path::PathType::Tag(_) => return Err(Error::from(ErrorKind::PermissionDenied))
		};
		let path = path.path;
		// TODO:  Differentiating between a physical document and a symlinked
		//    document should probably be done in parse_path() so we don't have
		//    to do these prefix checks twice.
		if path.strip_prefix("documents").is_ok() {
			Ok(user.delete_document(doc.get_id()).await?)
		} else if let Ok(path) = path.strip_prefix("/tags") {
			let tag = user.get_tag_by_canonical_path(path.parent().unwrap()).await?;
			user.untag_document(doc.get_id(), tag.get_id()).await?;
			Ok(())
		} else {
			Err(Error::from(ErrorKind::FileNameNotAllowedError))
		}
	} // }}}

	async fn rmd<P: AsRef<Path> + Send>(&self, user: &Session, path: P) -> Result<()> /* {{{ */ {
		let path = self.parse_path(user, path).await?;
		match path.contents {
			path::PathType::Root => Err(Error::from(ErrorKind::PermissionDenied)),
			path::PathType::Directory => Ok(user.rmdir(&path.internal_path).await?),
			path::PathType::Document(Some(_)) => Err(Error::from(ErrorKind::PermissionDenied)),
			path::PathType::Document(None) => Err(Error::from(ErrorKind::PermanentFileNotAvailable)),
			path::PathType::Tag(None) => Err(Error::from(ErrorKind::PermissionDenied)),
			path::PathType::Tag(Some(tag)) => Ok(user.delete_tag(tag.get_id()).await?)
		}
	} // }}}

	async fn mkd<P: AsRef<Path> + Send>(&self, user: &Session, path: P) -> Result<()> /* {{{ */ {
		let path = path.as_ref();
		self.parse_path(user, path).await?;
		let root = Utf8PathBuf::from("/");
		if let Ok(path) = path.strip_prefix("/documents") {
			let path: &Utf8Path = path.try_into().map_err(|e| Error::new(ErrorKind::FileNameNotAllowedError, e))?;
			let path = root.join(path);
			Ok(user.mkdir(&path).await?)
		} else if let Ok(path) = path.strip_prefix("/tags") {
			let path: &Utf8Path = path.try_into().map_err(|e| Error::new(ErrorKind::FileNameNotAllowedError, e))?;
			let path = root.join(path);
			let leaf = path.file_name().ok_or_else(|| Error::from(ErrorKind::FileNameNotAllowedError))?;
			let parent_id = match path.parent() {
				Some(parent) if parent == root => None,
				Some(parent) => Some(user.get_tag_by_canonical_path(parent).await?.get_id()),
				None => None
			};
			user.create_tag(parent_id, leaf.as_ref()).await?;
			Ok(())
		} else {
			Err(Error::from(ErrorKind::FileNameNotAllowedError))
		}
	} // }}}

	async fn rename<P: AsRef<Path> + Send>(&self, user: &Session, from: P, to: P) -> Result<()> /* {{{ */ {
		let from = self.parse_path(user, from).await?;
		let to = self.parse_path(user, to).await?;
		to.check_receive(&from)?;
		match from.contents {
			path::PathType::Directory => match to.contents {
				path::PathType::Directory | path::PathType::Document(None) => user.move_dir(&from.internal_path, &to.internal_path).await?,
				// Error should have already been thrown by to.check_receive()
				_ => return Err(Error::from(ErrorKind::LocalError))
			},
			path::PathType::Document(Some(doc)) => match to.contents {
				path::PathType::Directory => user.move_document(doc.get_id(), &to.internal_path).await?,
				path::PathType::Document(dest) => {
					if let Some(ddoc) = dest {
						user.delete_document(ddoc.get_id()).await?;
					}
					user.move_document(doc.get_id(), &to.internal_path).await?;
				},
				// Error should have already been thrown by to.check_receive()
				_ => return Err(Error::from(ErrorKind::LocalError))
			},
			// Error should have already been thown by to.check_receive()
			_ => return Err(Error::from(ErrorKind::LocalError))
		};
		Ok(())
	} // }}}

	async fn cwd<P: AsRef<Path> + Send>(&self, user: &Session, path: P) -> Result<()> /* {{{ */ {
		let path = self.parse_path(user, path).await?;
		match path.contents {
			path::PathType::Root => Ok(()),
			path::PathType::Directory => {
				let path = user.canonicalize_path(&path.internal_path)?;
				match tokio::fs::read_dir(path).await {
					Ok(_) => Ok(()),
					Err(e) => match e.kind() {
						std::io::ErrorKind::NotFound => Err(Error::from(ErrorKind::PermanentFileNotAvailable)),
						std::io::ErrorKind::PermissionDenied => Err(Error::from(ErrorKind::PermissionDenied)),
						_ => Err(Error::from(ErrorKind::LocalError))
					}
				}
			},
			path::PathType::Document(_) => Err(Error::from(ErrorKind::LocalError)),
			path::PathType::Tag(Some(_)) => Ok(()),
			path::PathType::Tag(None) => {
				if(path.path == "/tags") {
					Ok(())
				} else {
					Err(Error::from(ErrorKind::PermanentFileNotAvailable))
				}
			}
		}
	} // }}}
}

