extern crate yadms_storage;

use std::convert::TryFrom;

use camino::Utf8Path;
use camino::Utf8PathBuf;
use libunftp::storage::Error;
use libunftp::storage::ErrorKind;

use yadms_storage::document::Document;
use yadms_storage::tag::Tag;

use super::error;

#[derive(Debug)]
pub(crate) enum PathType {
	Root,
	Directory,
	Document(Option<Document>),
	Tag(Option<Tag>)
}

#[derive(Debug, PartialEq)]
pub(crate) enum Tree /* {{{ */ {
	Documents,
	Tags
}

impl TryFrom<&Utf8Path> for Tree {
	type Error = error::InvalidPathError;

	fn try_from(path: &Utf8Path) -> Result<Tree, Self::Error> {
		if(path.starts_with("/documents")) {
			Ok(Tree::Documents)
		} else if(path.starts_with("/tags")) {
			Ok(Tree::Tags)
		} else {
			Err(Self::Error::new(path))
		}
	}
}
// }}}

#[derive(Debug)]
pub(crate) struct Path {
	pub(crate) path: Utf8PathBuf,
	pub(crate) internal_path: Utf8PathBuf,
	pub(crate) contents: PathType
}

impl Path {
	pub(crate) fn new(path: Utf8PathBuf, internal_path: Utf8PathBuf, contents: PathType) -> Self /* {{{ */ {
		Self{
			path: path,
			internal_path: internal_path,
			contents: contents
		}
	} // }}}

	pub(crate) fn check_move(&self) -> Result<(), Error> /* {{{ */ {
		match &self.contents {
			PathType::Root => Err(Error::from(ErrorKind::PermissionDenied)),
			PathType::Directory => Ok(()),
			PathType::Document(Some(_doc)) => {
				if(self.path.strip_prefix("/tags").is_ok()) {
					// TODO:  Allow this
					Err(Error::from(ErrorKind::PermissionDenied))
				} else {
					Ok(())
				}
			},
			PathType::Document(None) => Err(Error::from(ErrorKind::PermanentFileNotAvailable)),
			PathType::Tag(Some(_tag)) => {
				//TODO:  Allow this
				Err(Error::from(ErrorKind::PermissionDenied))
			},
			PathType::Tag(None) => Err(Error::from(ErrorKind::PermanentFileNotAvailable))
		}
	} // }}}

	pub(crate) fn check_receive(&self, source: &Self) -> Result<(), Error> {
		source.check_move()?;
		match &self.contents {
			PathType::Root => Err(Error::from(ErrorKind::PermissionDenied)),
			PathType::Directory => match source.contents {
				PathType::Directory => Ok(()),
				PathType::Document(Some(_)) => Ok(()),
				_ => Err(Error::from(ErrorKind::PermissionDenied))
			},
			PathType::Document(Some(_doc)) => match &source.contents {
				PathType::Document(Some(_from_doc)) => Ok(()),
				_ => Err(Error::from(ErrorKind::PermissionDenied))
			},
			PathType::Document(None) => match &source.contents {
				PathType::Document(Some(_from_doc)) => Ok(()),
				PathType::Directory => Ok(()),
				_ => Err(Error::from(ErrorKind::PermissionDenied))
			}
			PathType::Tag(Some(_tag)) => {
				// TODO:  Allow this
				Err(Error::from(ErrorKind::PermissionDenied))
			}
			PathType::Tag(None) => Err(Error::from(ErrorKind::PermanentFileNotAvailable)),
		}
	}
}

