import hashlib
import io
import os
import threading
import time
import unittest

import ftplib
import pymysql

def query(sql, *args): # {{{
	db = pymysql.connect(
		host = os.getenv('MYSQL_HOST'),
		port = int(os.getenv('MYSQL_PORT')),
		user = os.getenv('MYSQL_USERNAME'), 
		password = os.getenv('MYSQL_PASSWORD'),
		db = os.getenv('MYSQL_DATABASE')
	)
	cursor = db.cursor()
	cursor.execute(sql, *args)
	return cursor.fetchall()
# }}}

def get_searchtext(path): # {{{
	rows = query('SELECT searchtext FROM documents WHERE path = %s', path)
	return rows[0][0]
# }}}

def upload_file_binary(ftp, path, remote_path): # {{{
	with open(os.path.join(os.path.dirname(__file__), path), 'rb') as f:
		return ftp.storbinary('STOR %s' % (remote_path,), f)
# }}}

def upload_file_ascii(ftp, path, remote_path): # {{{
	with open(os.path.join(os.path.dirname(__file__), path), 'r') as f:
		return ftp.storlines('STOR %s' % (remote_path,), f)
# }}}

def md5_file(path): # {{{
	hasher = hashlib.md5()
	with open(path, 'rb') as f:
		hasher.update(f.read())
	return hasher.hexdigest()
# }}}

def sorted(arr): # {{{
	arr.sort()
	return arr
# }}}

ftp = ftplib.FTP()
ftp.connect('localhost', port = 5021)
ftp.set_pasv(True)
ftp.login() # TODO:  Authenticate, once auth is supported by YADMS

# Expected base state {{{
assert(ftp.pwd() == '/')
# TODO:  [v for v in ftp.mlsd()] # once MLST/MLSD are supported in libunftp
assert(ftp.nlst() == ['documents', 'tags'])
assert(ftp.cwd('tags') == '250 Successfully cwd')
assert(ftp.pwd() == '/tags')
assert(ftp.nlst() == [])
# TODO:  ftp.cwd('../documents') when (if?) libunftp resolves relative paths to
#    absolute for cwd.  The way it works now functions fine, but keeps
#    appending .. to the path rather than resolving it.
assert(ftp.cwd('/documents') == '250 Successfully cwd')
assert(ftp.pwd() == '/documents')
assert(ftp.nlst() == [])
assert(ftp.cwd('/') == '250 Successfully cwd')
assert(ftp.pwd() == '/')
# }}}

# Create a directory and a couple documents {{{
assert(ftp.cwd('documents') == '250 Successfully cwd')
assert(ftp.pwd() == '/documents')
assert(ftp.mkd('derp') == '')
assert(ftp.mkd('herp') == '')
with unittest.TestCase.assertRaises(None, ftplib.error_temp):
	ftp.mkd('herp')
assert(ftp.storlines("STOR herp/lipsum.txt", io.BytesIO(b"lorem ipsum dolor sit amet\nsecond line\n")) == '226 File successfully written')
assert(get_searchtext('/herp/lipsum.txt') == 'lorem ipsum dolor sit amet second line ') # TODO:  rm trailing space
assert(md5_file('docs/herp/lipsum.txt') == '8271ac2a2111aebcf6d4083e038a2d4e')
assert(sorted(ftp.nlst()) == ['derp', 'herp'])
assert(sorted(ftp.nlst('/documents')) == ['derp', 'herp'])
assert(ftp.nlst('derp') == [])
assert(ftp.nlst('herp') == ['lipsum.txt'])
assert(upload_file_binary(ftp, 'lipsum-color.pdf', 'derp/lipsum-color.pdf') == '226 File successfully written')
#assert(get_searchtext('/derp/lipsum-color.pdf') == 'Lorem Ipsum All the facts Lipsum generator https wwwlipsumcomfeedhtml Lovem Ipsum Neque porro quisquam est qui dolorem ipsurm quia dolor sit amet consectetur adipisci velit There is no one who loves pain itself who seeks after it and wants to have it simply because itis pain Lorem ipsum dolor sit amet consectetur adipiscing elit Nunc commodo in purus a pulvinar Cras cursus non nulla et pretium Proin ac malesuada tortor sit amet ornare eros Cras in ex a elit iaculis egestas ac ut justo Aliquam erat volutpat Nam et sem efficitur ornare ligula eu luctus ante Suspendisse leo enim ultrices quis laoreet ac dignissim non augue Morbi in libero at nulla maximus pretium Nulla facilisi Nunc consectetur magna sit amet efficitur pulvinar Vestibulum molestie felis sed vehicula laoreet risus nisi mattis mauris id convallis ex arcu sit amet dui Quisque eget eleifend elit vitae accumsan mauris Aliquam lorem orci ullamcorper eu nunc eu placerat ullamcorper ipsum Generated 1 paragraph 110 words 708 bytes of Lorem Ipsum helplipsumcom Privacy Policy')
# Apparently tesseract (or maybe just ocrmypdf?) doesn't produce files with
#    completely deterministic contents.  TODO:  Can we figure out a way to
#    validate this?
#assert(md5_file('docs/derp/lipsum-color.pdf') == 'da41fc9ee89a14b17d5b56aa016a7060')
assert(ftp.cwd('derp') == '250 Successfully cwd')
assert(upload_file_binary(ftp, 'lipsum-grayscale.pdf', 'lipsum-grayscale.pdf') == '226 File successfully written')
#assert(get_searchtext('/derp/lipsum-grayscale.pdf') == 'Lorem Ipsum All the facts Lipsum generator https wwwlipsumcomfeedhtml Lovem Ipsum Neque porro quisquam est qui dolorem ipsum quia dolor sit amet consectetur adipisci velit There is no one who loves pain itself who seeks after it and wants to have it simply because it is pain Lorem ipsum dolor sit amet consectetur adipiscing elit Nunc commodo in purus a pulvinar Cras cursus non nulla et pretium Proin ac malesuada tortor sit amet ornare eros Cras in ex a elit iaculis egestas ac ut justo Aliquam erat volutpat Nam et sem efficitur ornare ligula eu luctus ante Suspendisse leo enim ultrices quis laoreet ac dignissim non augue Morbi in libero at nulla maximus pretium Nulla facilisi Nunc consectetur magna sit amet efficitur pulvinar Vestibulum molestie felis sed vehicula laoreet risus nisi mattis mauris id convallis ex arcu sit amet dui Quisque eget eleifend elit vitae accumsan mauris Aliquam lorem orci ullamcorper eu nunc eu placerat ullamcorper ipsum Generated 1 paragraph 110 words 708 bytes of Lorem Ipsum helplipsumcom Privacy Policy')
#assert(md5_file('docs/derp/lipsum-grayscale.pdf') == '504c2009f9bfe90ed7970a6eb6c53684')
assert(ftp.cwd('..') == '250 OK') # TODO:  Why is it '250 OK' instead of '250 Successfully cwd'?
assert(upload_file_binary(ftp, 'lipsum-bnw.pdf', 'derp/lipsum-bnw.pdf') == '226 File successfully written')
#assert(get_searchtext('/derp/lipsum-bnw.pdf') == 'Lorem Ipsum All the facts Lipsum generator https wwwlipsumcomfeedhtml Lovem Ipsum Neque porro quisquam est qui dolorem ipsumn quia dolor sit amet consectetur adipisci velit There is no one who loves pain itself who seeks after it and wants to have it simply because it is pain Lorem ipsum dolor sit amet consectetur adipiscing elit Nunc commodo in purus a pulvinar Cras cursus non nulla et pretium Proin ac malesuada tortor sit amet ornare eros Cras in ex a elit iaculis egestas ac ut justo Aliquam erat volutpat Nam et sem efficitur ornare ligula eu luctus ante Suspendisse leo enim ultrices quis laoreet ac dignissim non augue Morbi in libero at nulla maximus pretium Nulla facilisi Nunc consectetur magna sit amet efficitur pulvinar Vestibulum molestie felis sed vehicula laoreet risus nisi mattis mauris id convallis ex arcu sit amet dui Quisque eget eleifend elit vitae accumsan mauris Aliquam lorem orci ullamcorper eu nunc eu placerat ullamcorper ipsum Generated 1 paragraph 110 words 708 bytes of Lorem Ipsum helplipsumcom Privacy Policy')
#assert(md5_file('docs/derp/lipsum-bnw.pdf') == 'caa6b3bd527a1f40b2df8cb60f924884')
assert(sorted(ftp.nlst()) == ['derp', 'herp'])
assert(sorted(ftp.nlst('derp')) == ['lipsum-bnw.pdf', 'lipsum-color.pdf', 'lipsum-grayscale.pdf'])
assert(ftp.nlst('herp') == ['lipsum.txt'])
# }}}

# Create some tags {{{
assert(ftp.cwd('/tags') == '250 Successfully cwd')
assert(ftp.mkd('depth1tag1') == '')
assert(ftp.mkd('depth1tag2') == '')
assert(ftp.mkd('depth1tag3') == '')
assert(ftp.mkd('depth1tag1/depth2tag1') == '')
assert(ftp.mkd('depth2tag2') == '')
assert(ftp.mkd('depth2tag3') == '')
assert(ftp.mkd('depth1tag1/depth2tag1/depth3tag1') == '')
assert(ftp.mkd('depth1tag1/depth2tag1/depth3tag2') == '')
assert(ftp.mkd('depth3tag3') == '')
assert(sorted(ftp.nlst()) == ['depth1tag1', 'depth1tag2', 'depth1tag3', 'depth2tag2', 'depth2tag3', 'depth3tag3'])
assert(ftp.nlst('depth1tag1') == ['depth2tag1'])
assert(ftp.nlst('depth1tag2') == [])
assert(ftp.nlst('depth1tag3') == [])
assert(sorted(ftp.nlst('depth1tag1/depth2tag1')) == ['depth3tag1', 'depth3tag2'])
# TODO:  Tag some docs
# }}}

# Move stuff {{{
assert(ftp.cwd('/documents') == '250 Successfully cwd')
assert(ftp.rename('derp', 'herp/pdfs') == '250 Renamed')
assert(ftp.nlst() == ['herp'])
assert(sorted(ftp.nlst('herp')) == ['lipsum.txt', 'pdfs'])
assert(sorted(ftp.nlst('herp/pdfs')) == ['lipsum-bnw.pdf', 'lipsum-color.pdf', 'lipsum-grayscale.pdf'])
assert(ftp.cwd('herp/pdfs') == '250 Successfully cwd')
assert(ftp.rename('lipsum-bnw.pdf', '..') == '250 Renamed')
assert(sorted(ftp.nlst()) == ['lipsum-color.pdf', 'lipsum-grayscale.pdf'])
assert(ftp.cwd('..') == '250 OK')
assert(sorted(ftp.nlst()) == ['lipsum-bnw.pdf', 'lipsum.txt', 'pdfs'])
assert(ftp.cwd('/documents/herp') == '250 Successfully cwd')
assert(sorted(ftp.nlst()) == ['lipsum-bnw.pdf', 'lipsum.txt', 'pdfs'])
assert(ftp.cwd('/tags') == '250 Successfully cwd')
# TODO:  Moving tags is not currently supported; once it is, enable these
#assert(ftp.rename('depth3tag3', 'depth2tag3/depth3tag3') == '250 Renamed')
#assert(ftp.rename('depth2tag3', 'depth1tag2/depth2tag3') == '250 Renamed')
#assert(sorted(ftp.nlst()) == ['depth1tag1', 'depth1tag2', 'depth1tag3'])
# }}}

# Deletion {{{
assert(ftp.cwd('/documents/herp') == '250 Successfully cwd')
with unittest.TestCase.assertRaises(None, ftplib.error_temp):
	# 451 Local error
	ftp.rmd('pdfs')
with unittest.TestCase.assertRaises(None, ftplib.error_perm):
	# 550 Permission denied
	# (Yes, "permission denied" for trying to delete a directory as if it's a file.)
	ftp.delete('pdfs')
assert(ftp.delete('lipsum.txt') == '250 File successfully removed')
assert(ftp.delete('pdfs/lipsum-color.pdf') == '250 File successfully removed')
assert(sorted(ftp.nlst()) == ['lipsum-bnw.pdf', 'pdfs'])
assert(ftp.nlst('pdfs') == ['lipsum-grayscale.pdf'])
assert(ftp.cwd('pdfs') == '250 Successfully cwd')
assert(ftp.nlst() == ['lipsum-grayscale.pdf'])
assert(ftp.delete('lipsum-grayscale.pdf') == '250 File successfully removed')
assert(ftp.nlst() == [])
assert(sorted(ftp.nlst('..')) == ['lipsum-bnw.pdf', 'pdfs'])
assert(ftp.delete('../lipsum-bnw.pdf') == '250 File successfully removed')
assert(ftp.cwd('..') == '250 OK')
assert(ftp.nlst() == ['pdfs'])
assert(ftp.rmd('pdfs') == '250 File successfully removed')
assert(ftp.nlst() == [])
assert(ftp.cwd('/tags'))
assert(sorted(ftp.nlst()) == ['depth1tag1', 'depth1tag2', 'depth1tag3', 'depth2tag2', 'depth2tag3', 'depth3tag3'])
# TODO:  Fix this then enable the test
#with unittest.TestCase.assertRaises(None, ftplib.error_temp):
#	# 451 Local error
#	ftp.rmd('depth1tag1')
with unittest.TestCase.assertRaises(None, ftplib.error_perm):
	# 550 Permission denied
	ftp.delete('depth3tag3')
# }}}

# Upload a file, but with time delays in the stream {{{
class BytesLoop: # {{{
	def __init__(self, data = b''):
		self.buffer = data
		self.is_closed = False
	def read(self, count = -1):
		if(count == -1):
			chunk = self.buffer
			self.buffer = b''
			return chunk
		elif(count > len(self.buffer)):
			count = len(self.buffer)
		chunk = self.buffer[:count]
		self.buffer = self.buffer[count:]
		return chunk
	def readline(self, max_len):
		while len(self.buffer) == 0 and not self.closed():
			time.sleep(0.1)
		i = self.buffer.find(b'\n', 0, max_len - 1)
		if(i == -1):
			i = max_len
		return self.read(i + 1)
	def write(self, data):
		self.buffer += data
	def close(self):
		self.is_closed = True
	def closed(self):
		return self.is_closed
# }}}

def write_lipsum_with_delays(buf): # {{{
	buf.write(b'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Curabitur maximus est et mi pulvinar, interdum elementum velit consectetur. Duis ultrices ac lacus sed sagittis. Integer fringilla mauris ex, et sodales libero sagittis ut. Mauris in accumsan quam. Mauris nec velit auctor, hendrerit tellus volutpat, rhoncus neque. Sed molestie porta dignissim. Mauris tristique tortor eget ullamcorper volutpat. Aliquam finibus efficitur odio, in maximus justo cursus eget.\n\n')
	buf.write(b'Pellentesque iaculis nisi dolor, non ultrices ligula ullamcorper pellentesque. Maecenas porta elementum purus id elementum. Nullam eget ligula feugiat, fringilla mi eu, dictum leo. Vivamus malesuada augue sit amet dui lobortis vestibulum. Fusce placerat placerat orci, eget semper ante tempor ac. Nam imperdiet felis et augue scelerisque, at viverra lorem vulputate. Donec rutrum rhoncus libero, in aliquet dolor fermentum convallis. Nullam dolor orci, molestie id tellus vel, pulvinar eleifend risus. Cras vitae congue mi, id scelerisque turpis. Aenean volutpat erat quis tortor efficitur gravida. Pellentesque nec arcu aliquet, sodales diam eu, sollicitudin felis. Fusce id quam a sapien luctus condimentum vehicula vel nisl. Nam velit tortor, maximus vitae justo et, sagittis euismod ex.\n\n')
	time.sleep(1)
	buf.write(b'Nunc venenatis enim sed faucibus tempus. Curabitur rutrum ligula et vulputate euismod. In pharetra euismod sem sit amet rutrum. Donec eu venenatis sem, quis luctus lacus. Donec volutpat elit erat, vel rhoncus massa rhoncus id. Cras eget efficitur justo, eget ultricies dui. Donec rhoncus nibh libero, non fermentum purus ornare ut. Nam ullamcorper ex vitae risus lobortis commodo. Vestibulum ante ipsum primis in faucibus orci luctus et ultrices posuere cubilia curae; Sed tincidunt arcu ac massa iaculis imperdiet. Donec consequat, dolor a venenatis dapibus, eros urna gravida ex, in fringilla mauris quam a ligula. Nam consequat eros eu dictum venenatis. Donec finibus a arcu mattis accumsan. Curabitur varius nulla egestas neque dictum, vel cursus mauris consequat. Ut sagittis, dolor vitae ultricies porttitor, orci leo aliquet tortor, id aliquet neque felis sit amet augue.\n\n')
	time.sleep(3)
	buf.write(b'Sed cursus, neque ac laoreet consectetur, ipsum ligula suscipit nulla, et laoreet magna elit at dolor. Donec vel maximus sapien, at pretium sapien. Quisque eget maximus lectus. Etiam placerat, risus at posuere interdum, augue mauris tempus risus, sit amet sodales leo tellus a eros. Cras eu urna egestas, imperdiet magna vitae, sagittis erat. Etiam laoreet nisi justo. Praesent mattis, purus sed commodo sagittis, turpis ligula laoreet lectus, molestie volutpat nibh felis sit amet ex. Pellentesque nec luctus diam, vitae feugiat enim. Aliquam ut odio est. Praesent auctor magna id enim tristique maximus. Praesent scelerisque dui non enim posuere, at egestas risus viverra. Phasellus eu bibendum risus. Praesent a varius augue.\n\n')
	time.sleep(9)
	buf.write(b'Aliquam convallis enim at mauris maximus, eu sollicitudin metus fringilla. Maecenas placerat, urna vel gravida fermentum, lectus diam dictum odio, eu consequat quam ipsum nec orci. Phasellus a fermentum arcu. Suspendisse ac mi euismod, interdum tellus a, dignissim ipsum. Aliquam aliquet nulla orci, vel viverra nisi blandit at. Donec ornare nulla non dolor facilisis, mattis tincidunt est sagittis. Nulla molestie, lorem vel hendrerit porta, nunc metus dapibus augue, vel rhoncus magna turpis nec velit. Mauris dui est, semper sed arcu scelerisque, sagittis bibendum lorem. Etiam condimentum elit cursus mauris efficitur dapibus eu ut est. Praesent feugiat orci et magna laoreet, et porttitor lorem facilisis. Vestibulum sit amet fermentum velit, vel efficitur odio. Suspendisse velit nulla, molestie a fermentum vitae, tincidunt pellentesque diam. Nunc leo mauris, consectetur vel tincidunt at, tincidunt eu ipsum.\n\n')
	time.sleep(27)
	buf.write(b'Aliquam efficitur, turpis ac pellentesque tempus, odio nunc mattis ligula, sed rutrum eros sem sed lacus. Sed id mauris in tortor volutpat varius. Morbi vitae consequat neque. Pellentesque accumsan aliquam turpis vel pharetra. Suspendisse aliquet lacus tristique sodales vehicula. Nunc ante nisi, ornare vel ex vel, convallis cursus ex. Suspendisse orci lorem, auctor vel ex nec, consectetur volutpat tortor. Nullam vel tempor quam. Class aptent taciti sociosqu ad litora torquent per conubia nostra, per inceptos himenaeos. Pellentesque quis lobortis lorem. Morbi at pharetra dolor, id lacinia mi. Cras cursus est enim, at euismod turpis hendrerit ut. Nulla nec nisi hendrerit, finibus elit tincidunt, molestie urna.\n\n')
	time.sleep(81)
	buf.write(b'Donec quis est eu odio ultricies dictum. Donec ante arcu, blandit id lacus sit amet, commodo tincidunt erat. Quisque dictum augue sapien, quis pharetra sapien faucibus nec. Proin rhoncus lectus sed ullamcorper scelerisque. Proin in eros sodales, viverra ligula a, elementum nisi. Quisque facilisis nisl vitae dui mattis accumsan. Duis vulputate tortor felis, eget scelerisque sapien auctor nec. Lorem ipsum dolor sit amet, consectetur adipiscing elit. Aenean gravida massa aliquam facilisis elementum. Quisque accumsan lacus dolor, in dapibus neque maximus at. Nunc erat ligula, pellentesque at quam ac, eleifend finibus diam. Mauris turpis arcu, placerat sit amet diam et, commodo viverra velit.\n\n')
	time.sleep(27)
	buf.close()
# }}}

assert(ftp.cwd('/documents') == '250 Successfully cwd')
buf = BytesLoop()
t = threading.Thread(target = write_lipsum_with_delays, args = (buf,))
t.start()
ftp.storlines('STOR herp/delayed_lines.txt', buf, print)
t.join()
# }}}

assert(ftp.quit() == '221 Bye!')

